%{
/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

/*************************************************************************/
/* NOTE: If you add an #include here, make sure you properly update the  */
/*       parser.o and parser.po dependency lines in the Makefile.        */
/*************************************************************************/

#include <ctype.h>

#include "bytecodes.h"
#include "config.h"
#include "functions.h"
#include "opcode.h"
#include "parser.h"
#include "program.h"
#include "storage.h"
#include "streams.h"
#include "structures.h"
#include "sym_table.h"
#include "utils.h"
#include "vector.h" 

static Stmt     	       *prog_start;

static void			error(const char *, const char *);
static struct Expr_Ident	find_id(const char *name);
static void			yyerror(const char *s);
static int			yylex(void);
%}

%union {
  Stmt           *stmt;
  Expr           *expr;
  int             number;
  const char     *string;
  enum error      error;
  Arg_List       *args;
  Cond_Arm       *arm;
}

%type	<stmt> statements statement elsepart 
%type	<arm>  elseifs
%type   <expr> expr
%type   <args> arglist ne_arglist

%token	<number> NUMBER
%token	<string> STRING ID
%token	<error> ERROR
%token	IF ELSE ELSEIF ENDIF FORLIST FOR IN ENDFOR RETURN
%token	FORK ENDFORK WHILE ENDWHILE

%token	TO

%right	'='
%nonassoc '?' '|'
%left	OR AND
%left   EQ NE LT LE GT GE IN
%left	'+' '-'
%left	'*' '/' '%'
%left	NOT UNARYMINUS
%nonassoc '.' ':' '[' '$'

%%

program:  statements
		{ prog_start = $1; }
	;

statements:
	  /* NOTHING */
		{ $$ = 0; }
	| statements statement
		{
		    if ($1) {
			Stmt *tmp = $1;
			
			while (tmp->next)
			    tmp = tmp->next;
			tmp->next = $2;
			$$ = $1;
		    } else
			$$ = $2;
		}
	;

statement:
	  IF '(' expr ')' statements elseifs elsepart ENDIF
		{

		    $$ = alloc_stmt(STMT_COND);
		    $$->s.cond.arms = alloc_cond_arm($3, $5);
		    $$->s.cond.arms->next = $6;
		    $$->s.cond.otherwise = $7;
		}
	| FOR ID IN '(' expr ')' statements ENDFOR
		{
		    $$ = alloc_stmt(STMT_LIST);
		    $$->s.list.id = find_id($2);
		    $$->s.list.expr = $5;
		    $$->s.list.body = $7;
		}
	| FOR ID IN '[' expr TO expr ']' statements ENDFOR
		{
		    $$ = alloc_stmt(STMT_RANGE);
		    $$->s.range.id = find_id($2);
		    $$->s.range.from = $5;
		    $$->s.range.to = $7;
		    $$->s.range.body = $9;
		}
	| WHILE '(' expr ')' statements ENDWHILE
		{
		    $$ = alloc_stmt(STMT_WHILE);
		    $$->s.loop.condition = $3;
		    $$->s.loop.body = $5;
		}
	| FORK '(' expr ')' statements ENDFORK
		{
		    $$ = alloc_stmt(STMT_FORK);
		    $$->s.fork.id.name = 0;
		    $$->s.fork.time = $3;
		    $$->s.fork.body = $5;
		}
	| FORK ID '(' expr ')' statements ENDFORK
		{
		    $$ = alloc_stmt(STMT_FORK);
		    $$->s.fork.id = find_id($2);
		    $$->s.fork.time = $4;
		    $$->s.fork.body = $6;
		}
	| expr ';'
		{
		    $$ = alloc_stmt(STMT_EXPR);
		    $$->s.expr = $1;
		}
	| RETURN expr ';'
		{
		    $$ = alloc_stmt(STMT_RETURN);
		    $$->s.expr = $2;
		}
	| RETURN ';'
		{
		    $$ = alloc_stmt(STMT_RETURN);
		    $$->s.expr = 0;
		}
	| ';'
		{ $$ = 0; }
	;

elseifs:
	  /* NOTHING */
		{ $$ = 0; }
	| elseifs ELSEIF '(' expr ')' statements
		{
		    Cond_Arm *this_arm = alloc_cond_arm($4, $6);
		    
		    if ($1) {
		        Cond_Arm *tmp = $1;

			while (tmp->next)
			    tmp = tmp->next;
			tmp->next = this_arm;
			$$ = $1;
		    } else
			$$ = this_arm;
		}
	;

elsepart:
	  /* NOTHING */
		{ $$ = 0; }
	| ELSE statements
		{ $$ = $2; }
	;

expr:
	  NUMBER
		{
		    $$ = alloc_var(TYPE_NUM);
		    $$->e.var.v.num = $1;
		}
	| STRING
		{
		    $$ = alloc_var(TYPE_STR);
		    $$->e.var.v.str = $1;
		}
	| '#' NUMBER
		{
		    $$ = alloc_var(TYPE_OBJ);
		    $$->e.var.v.obj = $2;
		}
	| '#' '-' NUMBER
		{
		    $$ = alloc_var(TYPE_OBJ);
		    $$->e.var.v.obj = -($3);
		}
	| ERROR
		{
		    $$ = alloc_var(TYPE_ERR);
		    $$->e.var.v.err = $1;
		}
	| ID
		{
		    $$ = alloc_expr(EXPR_ID);
		    $$->e.id = find_id($1);
		}
	| '$' ID
		{
		    /* Treat $foo like #0.("foo") */
		    Expr *obj = alloc_var(TYPE_OBJ);
		    Expr *prop = alloc_var(TYPE_STR);
		    obj->e.var.v.obj = 0;
		    prop->e.var.v.str = $2;
		    $$ = alloc_binary(EXPR_PROP, obj, prop);
		}
	| expr '.' ID
		{
		    /* Treat foo.bar like foo.("bar") for simplicity */
		    Expr *prop = alloc_var(TYPE_STR);
		    prop->e.var.v.str = $3;
		    $$ = alloc_binary(EXPR_PROP, $1, prop);
		}
	| expr '.' '(' expr ')'
    		{
		    $$ = alloc_binary(EXPR_PROP, $1, $4);
		}
	| expr ':' ID '(' arglist ')'
		{
		    /* treat foo:bar(args) like foo:("bar")(args) */
		    Expr *verb = alloc_var(TYPE_STR);
		    verb->e.var.v.str = $3;
		    $$ = alloc_verb($1, verb, $5);
		}
	| expr ':' '(' expr ')' '(' arglist ')'
		{
		    $$ = alloc_verb($1, $4, $7);
		}
	| expr '[' expr ']'
		{   $$ = alloc_binary(EXPR_INDEX, $1, $3);
		}
	| expr '[' expr TO expr ']'
		{

		    $$ = alloc_expr(EXPR_RANGE);
		    $$->e.range.base = $1;
		    $$->e.range.from = $3;
		    $$->e.range.to = $5;
		}
	| expr '=' expr
                {
		    Expr *e = $1;

		    while (e)
			switch (e->kind) {
			  case EXPR_ID:
			    e = 0;
			    break;
			  case EXPR_PROP:
			    e = 0;
			    break;
			  case EXPR_INDEX:
			    e = e->e.bin.lhs;
			    break;
			  case EXPR_RANGE:
			    if (e == $1)
				e = e->e.range.base;
			    else {
				e = 0;
				yyerror("Illegal expression on left side of assignment.");
			    }
			    break;
			  default:
			    e = 0;
			    yyerror("Illegal expression on left side of assignment.");
			}
			    
		    switch ($1->kind) {
		      case EXPR_ID:
			$$ = alloc_asgn(ASGN_VAR, $3);
			$$->e.asgn.a.id = $1->e.id;
			break;
		      case EXPR_PROP:
			$$ = alloc_asgn(ASGN_PROP, $3);
			$$->e.asgn.a.bin = $1->e.bin;
			break;
		      case EXPR_INDEX:
			$$ = alloc_asgn(ASGN_REF, $3);
			$$->e.asgn.a.bin = $1->e.bin;
			break;
		      case EXPR_RANGE:
			$$ = alloc_asgn(ASGN_RANGE, $3);
			$$->e.asgn.a.range = $1->e.range;
			break;
		      default:
			$$ = $3; /* Doesn't matter what is assigned here. */
			         /* Just need to maintain integrity of tree */
		    }
		    myfree($1, M_AST);

	        }
	| ID '(' arglist ')'
		{
		    unsigned f_no;

		    $$ = alloc_expr(EXPR_CALL);
		    if ((f_no = number_func_by_name($1)) == FUNC_NOT_FOUND) {
			error("Unknown built-in function: ", $1);
			$$->e.call.func = 0;
		    } else
			$$->e.call.func = f_no; /*f->func;*/
		    $$->e.call.args = $3;
		    free_str((void *) $1);
		}
	| expr '+' expr
		{
		    $$ = alloc_binary(EXPR_PLUS, $1, $3);
		}
	| expr '-' expr
		{
		    $$ = alloc_binary(EXPR_MINUS, $1, $3);
		}
	| expr '*' expr
		{
		    $$ = alloc_binary(EXPR_TIMES, $1, $3);
		}
	| expr '/' expr
		{
		    $$ = alloc_binary(EXPR_DIVIDE, $1, $3);
		}
	| expr '%' expr
		{
		    $$ = alloc_binary(EXPR_MOD, $1, $3);
		}
	| expr AND expr
		{
		    $$ = alloc_binary(EXPR_AND, $1, $3);
		}
	| expr OR expr
		{
		    $$ = alloc_binary(EXPR_OR, $1, $3);
		}
	| expr EQ expr
		{
		    $$ = alloc_binary(EXPR_EQ, $1, $3);
		}
	| expr NE expr
		{
		    $$ = alloc_binary(EXPR_NE, $1, $3);
		}
	| expr LT expr
		{
		    $$ = alloc_binary(EXPR_LT, $1, $3);
		}
	| expr LE expr
		{
		    $$ = alloc_binary(EXPR_LE, $1, $3);
		}
	| expr GT expr
		{
		    $$ = alloc_binary(EXPR_GT, $1, $3);
		}
	| expr GE expr
		{
		    $$ = alloc_binary(EXPR_GE, $1, $3);
		}
	| expr IN expr
		{
		    $$ = alloc_binary(EXPR_IN, $1, $3);
		}
	| '-' expr  %prec UNARYMINUS
		{
		    if ($2->kind == EXPR_VAR && $2->e.var.type == TYPE_NUM) {
			$2->e.var.v.num = -$2->e.var.v.num;
		        $$ = $2;
		    } else {
			$$ = alloc_expr(EXPR_NEGATE);
			$$->e.expr = $2;
		    }
		}
	| NOT expr
		{
		    $$ = alloc_expr(EXPR_NOT);
		    $$->e.expr = $2;
		}
	| '(' expr ')'
		{ $$ = $2; }
	| '{' arglist '}'
		{
		    $$ = alloc_expr(EXPR_LIST);
		    $$->e.list = $2;
		}
	| expr '?' expr '|' expr
		{
		    $$ = alloc_expr(EXPR_COND);
		    $$->e.cond.condition = $1;
		    $$->e.cond.consequent = $3;
		    $$->e.cond.alternate = $5;
		}
	;

arglist:
	  /* NOTHING */
		{ $$ = 0; }
        | ne_arglist
                { $$ = $1; }

ne_arglist:
	expr
		{ $$ = alloc_arg_list(ARG_NORMAL, $1); }
	| '@' expr
		{ $$ = alloc_arg_list(ARG_SPLICE, $2); }
	| ne_arglist ',' expr
		{
		    Arg_List *this_arg = alloc_arg_list(ARG_NORMAL, $3);

		    if ($1) {
			Arg_List *tmp = $1;

			while (tmp->next)
			    tmp = tmp->next;
			tmp->next = this_arg;
			$$ = $1;
		    } else
			$$ = this_arg;
		}
	| ne_arglist ',' '@' expr
		{
		    Arg_List *this_arg = alloc_arg_list(ARG_SPLICE, $4);

		    if ($1) {
			Arg_List *tmp = $1;

			while (tmp->next)
			    tmp = tmp->next;
			tmp->next = this_arg;
			$$ = $1;
		    } else
			$$ = this_arg;
		}
	;

%%

static int		lineno, nerrors;
static Parser_Client	client;
static void	       *client_data;
static Names	       *local_names;

static struct Expr_Ident
find_id(const char *name)
{
    struct Expr_Ident	id;

    id.name = name;
    id.slot = find_or_add_name(&local_names, name);

    return id;
}

static void
yyerror(const char *s)
{
    error(s, 0);
}

static void
error(const char *s, const char *t)
{
    static Stream      *str = 0;

    if (str == 0)
	str = new_stream(100);
    nerrors++;
    if (t)
	stream_printf(str, "Line %d:  %s%s", lineno, s, t);
    else
	stream_printf(str, "Line %d:  %s", lineno, s);
    (*(client.error))(client_data, reset_stream(str));
}

static int
lex_getc(void)
{
    return (*(client.getch))(client_data);
}

static void
lex_ungetc(int c)
{
    (*(client.ungetch))(client_data, c);
}

static int
follow(int expect, int ifyes, int ifno)     /* look ahead for >=, etc. */
{
    int c = lex_getc();

    if (c == expect)
        return ifyes;
    lex_ungetc(c);
    return ifno;
}

struct tt_entry {
    const char	*name;
    int		token;
    enum error	error;
};

static struct tt_entry tokentable[] = {
    { "if",		IF },
    { "else",		ELSE },
    { "elseif",		ELSEIF },
    { "endif",		ENDIF },
    { "for",		FOR },
    { "in",		IN },
    { "endfor",		ENDFOR },
    { "fork",		FORK },
    { "endfork",	ENDFORK },
    { "return",		RETURN },
    { "while",		WHILE },
    { "endwhile",	ENDWHILE },

    { "E_NONE",		ERROR,	E_NONE },
    { "E_TYPE",		ERROR,	E_TYPE },
    { "E_DIV",		ERROR,	E_DIV },
    { "E_PERM",		ERROR,	E_PERM },
    { "E_PROPNF",	ERROR,	E_PROPNF },
    { "E_VERBNF",	ERROR,	E_VERBNF },
    { "E_VARNF",	ERROR,	E_VARNF },
    { "E_INVIND",	ERROR,	E_INVIND },
    { "E_RECMOVE",	ERROR,	E_RECMOVE },
    { "E_MAXREC",	ERROR,	E_MAXREC },
    { "E_RANGE",	ERROR,	E_RANGE },
    { "E_ARGS",		ERROR,	E_ARGS },
    { "E_NACC",		ERROR,	E_NACC },
    { "E_INVARG",	ERROR,	E_INVARG },
    { "E_QUOTA",	ERROR,	E_QUOTA },
};

static Stream  *token_stream = 0;

static int
yylex(void)
{
    int		c;

start_over:

    do {
	c = lex_getc();
	if (c == '\n') lineno++;
    } while (isspace(c));

    if (c == '/') {
	c = lex_getc();
	if (c == '*') {
	    for (;;) {
		c = lex_getc();
		if (c == '*') {
		    c = lex_getc();
		    if (c == '/')
			goto start_over;
		}
		if (c == EOF) {
		    yyerror("end of program while in a comment");
		    return c;
		}
	    }
	} else {
	    lex_ungetc(c);
	    return '/';
	}
    }

    if (isdigit(c)) {
	int	n = 0;
	do {
	    n = n * 10 + c - '0';
	    c = lex_getc();
	} while (isdigit(c));
	lex_ungetc(c);
	yylval.number = n;
	return NUMBER;
    }
    
    if (isalpha(c) || c == '_') {
	int	i;
	char	*buf;

	stream_add_char(token_stream, c);
	while (isalnum(c = lex_getc()) || c == '_')
	    stream_add_char(token_stream, c);
	lex_ungetc(c);
	buf = reset_stream(token_stream);

	/*
	 * check for multi-char alpha tokens (if, for, else, etc.)
	 */
	for (i = 0; i < Arraysize(tokentable, struct tt_entry); i++) {
	    if (!mystrcasecmp(buf, tokentable[i].name)) {
		int	t = tokentable[i].token;

		if (t == ERROR)
		    yylval.error = tokentable[i].error;
		return t;
	    }
	}
	yylval.string = ak_alloc_string(buf);
	return ID;
    }

    if (c == '"') {
	while(1) {
	    c = lex_getc();
	    if (c == '"')
		break;
	    if (c == '\\')
		c = lex_getc();
	    if (c == '\n' || c == EOF) {
		yyerror("missing quote");
		break;
	    }
	    stream_add_char(token_stream, c);
	}
	yylval.string = ak_alloc_string(reset_stream(token_stream));
	return STRING;
    }

    switch(c) {
      case '>':         return follow('=', GE, GT);
      case '<':         return follow('=', LE, LT);
      case '=':         return follow('=', EQ, '=');
      case '!':         return follow('=', NE, NOT);
      case '|':         return follow('|', OR, '|');
      case '&':         return follow('&', AND, '&');
      case '.':		return follow('.', TO, '.');
      default:          return c;
    }
}

Program *
parse_program(Parser_Client c, void *data)
{
    extern int	yyparse();
    
    Program *ret = 0;
    
    if (token_stream == 0)
	token_stream = new_stream(1024);
    nerrors = 0;
    lineno = 1;
    client = c;
    client_data = data;
    local_names = new_builtin_names();
    
    begin_code_allocation();
    yyparse();
    
    if (nerrors == 0)
	ret = new_program();
    
    if (nerrors > 0)
	return 0;
    
    /* NO ERRORS! HURRAY! */
    
    ret->fork_vectors_size = vectorize(prog_start, &ret->main_vector,
				       &ret->fork_vectors,
				       &ret->literals, &ret->num_literals);
    
    ret->num_var_names = local_names->size;
    ret->var_names = local_names->names;

    myfree(local_names, M_NAMES);
    free_stmt(prog_start);
    end_code_allocation(nerrors > 0);
    
    return ret;
}

char rcsid_parser[] = "$Id: parser.y,v 1.13 1992/10/23 23:03:47 pavel Exp $";

/* $Log: parser.y,v $
 * Revision 1.13  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.12  1992/10/23  22:21:43  pavel
 * Eliminated all uses of the useless macro NULL.
 *
 * Revision 1.11  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.10  1992/08/31  22:25:51  pjames
 * Changed some `char *'s to `const char *'
 * 
 * Revision 1.9  1992/08/28  23:18:38  pjames
 * Added ASGN_RANGE parsing code.  Fixed tiny memory leak.
 * 
 * Revision 1.8  1992/08/28  16:16:52  pjames
 * Changed `ak_dealloc_string()' to `free_str()'.
 * 
 * Revision 1.7  1992/08/12  01:49:58  pjames
 * Negative literals are now accepted (instead of negated positive literals).
 * 
 * Revision 1.6  1992/08/10  16:55:59  pjames
 * Updated #includes.
 * 
 * Revision 1.5  1992/07/30  21:22:08  pjames
 * Removed max_stack calculations (moved to vector.c).
 * 
 * Revision 1.4  1992/07/29  18:05:12  pjames
 * $$ no longer = NULL when there is an illegal left hand side of an
 * expression.
 * 
 * Revision 1.3  1992/07/27  18:12:41  pjames
 * Changed name of ct_env to var_names and const_env to literals.
 * Modified call to vectorize to use new argument format.
 * Assignment statements now check left hand side properly for allowable
 * expressions, and now frees memory from reused expressions.
 * 
 * Revision 1.2  1992/07/21  00:05:16  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 * 
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 */
